-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local g_thMain = g_thConversionSystem.mainEnvironment
THDensityMapConversionUtil = {}
local debugFlagId = THDebugUtil.createDebugFlag("THDensityMapConversionUtil")
THDensityMapConversionUtil.debugFlagId = debugFlagId
THDensityMapConversionUtil.extraFuncArgs = {
    cutFruitArea = {}
}
THDensityMapConversionUtil.conversionCache = {}
THDensityMapConversionUtil.functionCache = {}
THDensityMapConversionUtil.isDestructionInitialized = false
THDensityMapConversionUtil.NUM_TIP_LINES = {
    CHOPPER = 2
}
local g_thFarmlandManager = g_thConversionSystem.farmlandManager
local g_thFruitTypeManager = g_thConversionSystem.thFruitTypeManager
local g_thFillTypeManager = g_thConversionSystem.thFillTypeManager
local g_thPFManager = g_thConversionSystem.thPFManager
function THDensityMapConversionUtil.convertCutFruitArea(fruitType, stateId, sx, sz, wx, wz, hx, hz, toFruitType, toStateId, toGroundType, toChopperType, removeHaulm)
    local mission = g_thConversionSystem.mission
    local fieldGroundSystem = mission.fieldGroundSystem
    local terrainRootNode = THUtils.getNoNil(mission.terrainRootNode, 0)
    local fruitTypeConversionData = g_thConversionSystem:getFruitTypeConversionData(fruitType)
    local totalChangedArea, totalRealArea = 0, 0
    if fruitTypeConversionData ~= nil and terrainRootNode ~= 0 then
        local functionData = THDensityMapConversionUtil.functionCache.convertCutFruitArea
        if functionData == nil then
            functionData = {
                stateModifiers = {},
                stateFilters = {},
                preparingModifiers = {},
                preparingFilters = {}
            }
            local groundTypeMapId, groundTypeFirstChannel, groundTypeNumChannels = fieldGroundSystem:getDensityMapData(FieldDensityMap.GROUND_TYPE)
            if groundTypeMapId ~= nil then
                functionData.groundModifier = DensityMapModifier.new(groundTypeMapId, groundTypeFirstChannel, groundTypeNumChannels, terrainRootNode)
            end
            THDensityMapConversionUtil.functionCache.convertCutFruitArea = functionData
        end
        local fruitTypeInfo = fruitTypeConversionData.fruitType.info
        local toFruitTypeInfo = nil
        local fromState = -1
        local toState = -1
        if stateId ~= nil then
            fromState = g_thFruitTypeManager:getFruitTypeGrowthState(fruitTypeInfo.index, stateId, true)
            if fromState == nil or fromState < 0 then
                THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "stateId", stateId)
                fromState = -1 -- safety
            end
        end
        if toFruitType ~= nil then
            if toStateId == nil then
                THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "toStateId", toStateId)
            else
                toFruitTypeInfo = g_thFruitTypeManager:getFruitType(toFruitType)
                if toFruitTypeInfo == nil or toFruitTypeInfo.index == FruitType.UNKNOWN then
                    THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "toFruitType", toFruitType)
                    toFruitTypeInfo = nil -- safety
                end
            end
        else
            toFruitTypeInfo = fruitTypeInfo
        end
        if toStateId ~= nil and toFruitTypeInfo ~= nil then
            toState = g_thFruitTypeManager:getFruitTypeGrowthState(toFruitTypeInfo.index, toStateId, true)
            if toState == nil or toState < 0 then
                THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "toStateId", toStateId)
                toState = -1 -- safety
            end
        else
            toState = fromState
        end
        if functionData ~= nil and fromState >= 0 and toState >= 0
            and toFruitTypeInfo ~= nil and toFruitTypeInfo.index ~= FruitType.UNKNOWN
        then
            local fruitStateModifier = functionData.stateModifiers[toFruitTypeInfo.index]
            local fruitStateFilter = functionData.stateFilters[fruitTypeInfo.index]
            if fruitStateModifier == nil then
                local terrainDataPlaneId = toFruitTypeInfo.terrainDataPlaneId
                local startStateChannel = toFruitTypeInfo.startStateChannel
                local numStateChannels = toFruitTypeInfo.numStateChannels
                if terrainDataPlaneId ~= nil then
                    fruitStateModifier = DensityMapModifier.new(terrainDataPlaneId, startStateChannel, numStateChannels, terrainRootNode)
                    functionData.stateModifiers[toFruitTypeInfo.index] = fruitStateModifier
                end
            end
            if fruitStateFilter == nil then
                local terrainDataPlaneId = fruitTypeInfo.terrainDataPlaneId
                local startStateChannel = fruitTypeInfo.startStateChannel
                local numStateChannels = fruitTypeInfo.numStateChannels
                if terrainDataPlaneId ~= nil then
                    fruitStateFilter = DensityMapFilter.new(terrainDataPlaneId, startStateChannel, numStateChannels, terrainRootNode)
                    functionData.stateFilters[fruitTypeInfo.index] = fruitStateFilter
                end
            end
            if fruitStateFilter ~= nil then
                fruitStateFilter:setValueCompareParams(DensityValueCompareType.EQUAL, fromState)
                if toGroundType ~= nil then
                    local groundModifier = functionData.groundModifier
                    local groundTypeIndex = FieldGroundType.getByName(toGroundType)
                    local groundTypeValue = fieldGroundSystem:getFieldGroundValue(groundTypeIndex)
                    if groundModifier ~= nil and groundTypeValue ~= nil then
                        groundModifier:setParallelogramWorldCoords(sx, sz, wx, wz, hx, hz, DensityCoordType.POINT_POINT_POINT)
                        local changedArea, realArea = groundModifier:executeSet(groundTypeValue, fruitStateFilter)
                        if changedArea ~= nil and realArea ~= nil then
                            totalChangedArea = math.max(changedArea, totalChangedArea)
                            totalRealArea = math.max(realArea, totalRealArea)
                        end
                    end
                end
                if toChopperType ~= nil then
                end
                if removeHaulm == true then
                    local preparingModifier = functionData.preparingModifiers[fruitTypeInfo.index]
                    local preparingFilter = functionData.preparingFilters[fruitTypeInfo.index]
                    if preparingModifier == nil then
                        local terrainDataPlaneIdPreparing = fruitTypeInfo.terrainDataPlaneIdPreparing
                        preparingModifier = DensityMapModifier.new(terrainDataPlaneIdPreparing, 0, 1, terrainRootNode)
                        preparingFilter = DensityMapFilter.new(terrainDataPlaneIdPreparing, 0, 1, terrainRootNode)
                        preparingFilter:setValueCompareParams(DensityValueCompareType.GREATER, 0)
                        functionData.preparingModifiers[fruitTypeInfo.index] = preparingModifier
                        functionData.preparingFilters[fruitTypeInfo.index] = preparingFilter
                    end
                    if preparingModifier ~= nil then
                        preparingModifier:setParallelogramWorldCoords(sx, sz, wx, wz, hx, hz, DensityCoordType.POINT_POINT_POINT)
                        local changedArea, realArea = preparingModifier:executeSet(0, preparingFilter, fruitStateFilter)
                        if changedArea ~= nil and realArea ~= nil then
                            totalChangedArea = math.max(changedArea, totalChangedArea)
                            totalRealArea = math.max(realArea, totalRealArea)
                        end
                    end
                end
                if (toFruitTypeInfo ~= fruitTypeInfo or toState ~= fromState)
                    and fruitStateModifier ~= nil
                then
                    fruitStateModifier:setParallelogramWorldCoords(sx, sz, wx, wz, hx, hz, DensityCoordType.POINT_POINT_POINT)
                    local changedArea, realArea = fruitStateModifier:executeSet(toState, fruitStateFilter)
                    if changedArea ~= nil and realArea ~= nil then
                        totalChangedArea = math.max(changedArea, totalChangedArea)
                        totalRealArea = math.max(realArea, totalRealArea)
                    end
                end
            end
        end
    end
    return totalChangedArea, totalRealArea
end
function THDensityMapConversionUtil.getFruitTypeAreaModifiers(fruitType, sx, sz, wx, wz, hx, hz, useMinForageState, harvestParams)
    local mission = g_thConversionSystem.mission
    local fruitTypeConversionData = g_thConversionSystem:getFruitTypeConversionData(fruitType)
    local terrainRootNode = THUtils.getNoNil(mission.terrainRootNode, 0)
    if fruitTypeConversionData ~= nil and terrainRootNode ~= 0
        and harvestParams ~= nil
        and (harvestParams.inputFruitType == nil
            or harvestParams.inputFruitType == fruitTypeConversionData.fruitType.index)
    then
        local fruitTypeData = fruitTypeConversionData.fruitType.data
        local fruitTypeInfo = fruitTypeConversionData.fruitType.info
        local minHarvestState = fruitTypeData.minHarvestState
        local maxHarvestState = fruitTypeData.maxHarvestState
        local numGrowthStates = fruitTypeData.numGrowthStates
        if useMinForageState then
            minHarvestState = fruitTypeData.minForageState
        end
        local converterName = harvestParams.converterName
        local supportedFillTypes = harvestParams.supportedFillTypes
        local inputFillType = harvestParams.inputFillType
        local functionData = THDensityMapConversionUtil.functionCache.getFruitTypeAreaModifiers
        if functionData == nil then
            functionData = {
                fruitTypeModifiers = {},
                fruitTypeFilters = {}
            }
            THDensityMapConversionUtil.functionCache.getFruitTypeAreaModifiers = functionData
        end
        if numGrowthStates > 0 then
            local fruitTypeModifier = functionData.fruitTypeModifiers[fruitTypeInfo.index]
            local fruitTypeFilter = functionData.fruitTypeFilters[fruitTypeInfo.index]
            if fruitTypeModifier == nil or fruitTypeFilter == nil then
                local terrainDataPlaneId = fruitTypeInfo.terrainDataPlaneId
                local startStateChannel = fruitTypeInfo.startStateChannel
                local numStateChannels = fruitTypeInfo.numStateChannels
                if terrainDataPlaneId ~= nil and startStateChannel ~= nil
                    and numStateChannels ~= nil and numStateChannels > 0
                then
                    if fruitTypeModifier == nil then
                        fruitTypeModifier = DensityMapModifier.new(terrainDataPlaneId, startStateChannel, numStateChannels, terrainRootNode)
                        functionData.fruitTypeModifiers[fruitTypeInfo.index] = fruitTypeModifier
                    end
                    if fruitTypeFilter == nil then
                        if fruitTypeModifier ~= nil then -- only if modifier is valid
                            fruitTypeFilter = DensityMapFilter.new(fruitTypeModifier)
                            functionData.fruitTypeFilters[fruitTypeInfo.index] = fruitTypeFilter
                        end
                    end
                end
            end
            if fruitTypeModifier ~= nil and fruitTypeFilter ~= nil then
                fruitTypeModifier:setParallelogramWorldCoords(sx, sz, wx, wz, hx, hz, DensityCoordType.POINT_POINT_POINT)
                local finalModifierData = nil
                local usedGrowthState = -1
                local maxFruitArea = 0
                local mx = (wx + hx) / 2
                local mz = (wz + hz) / 2
                local moistureSystem = g_thMain:getModule("moistureSystem")
                local inputMoisture = -1
                local farmlandId = g_thFarmlandManager:getFarmlandIdAtWorldPosition(mx, mz)
                local soilTypeIndex = g_thPFManager:getSoilTypeIndexAtWorldPos(mx, mz, true) -- force
                for growthState = 1, numGrowthStates do
                    fruitTypeFilter:setValueCompareParams(DensityValueCompareType.EQUAL, growthState)
                    local _, fruitArea = fruitTypeModifier:executeGet(fruitTypeFilter)
                    if fruitArea ~= nil and fruitArea > maxFruitArea then
                        if moistureSystem ~= nil then
                            inputMoisture = moistureSystem:getFruitTypeMoisture(fruitTypeInfo.index, growthState, nil, farmlandId, soilTypeIndex)
                        end
                        local modifierData = g_thConversionSystem:getFruitTypeModifierData(fruitTypeInfo.index, growthState, converterName, inputFillType, inputMoisture, false, supportedFillTypes)
                        if (growthState >= minHarvestState and growthState <= maxHarvestState)
                            or (modifierData ~= nil and modifierData.values.toHarvestState ~= nil)
                        then
                            finalModifierData = modifierData
                            usedGrowthState = growthState
                            maxFruitArea = fruitArea
                        end
                    end
                end
                if usedGrowthState > 0 then
                    return finalModifierData, usedGrowthState
                end
            end
        end
    end
    return nil, -1
end
function THDensityMapConversionUtil.hook_updateWheelDestructionArea(superFunc, sx, sz, wx, wz, hx, hz, ...)
    if not THDensityMapConversionUtil.isDestructionInitialized then
        THUtils.call(function()
            local functionData = FSDensityMapUtil.functionCache.updateWheelDestructionArea
            if functionData == nil then
                return
            end
            local multiModifier = functionData.multiModifier
            local modifier = functionData.modifier
            local filter1 = functionData.filter1
            local filter2 = functionData.filter2
            if multiModifier ~= nil and modifier ~= nil and filter1 ~= nil and filter2 ~= nil then
                local fruitTypeList = g_thFruitTypeManager:getFruitTypeList()
                for _, fruitTypeInfo in pairs(fruitTypeList) do
                    local fruitTypeConversionData = g_thConversionSystem:getFruitTypeConversionData(fruitTypeInfo.index)
                    if fruitTypeConversionData ~= nil then
                        local destructionFilters = fruitTypeConversionData.destructionFilters
                        local numDestructionFilters = #destructionFilters
                        if numDestructionFilters > 0 then
                            local terrainDataPlaneId = fruitTypeInfo.terrainDataPlaneId
                            local startStateChannel = fruitTypeInfo.startStateChannel
                            local numStateChannels = fruitTypeInfo.numStateChannels
                            if terrainDataPlaneId ~= nil and startStateChannel ~= nil and numStateChannels ~= nil then
                                modifier:resetDensityMapAndChannels(terrainDataPlaneId, startStateChannel, numStateChannels)
                                filter1:resetDensityMapAndChannels(terrainDataPlaneId, startStateChannel, numStateChannels)
                                for filterIndex = 1, numDestructionFilters do
                                    local filterData = destructionFilters[filterIndex]
                                    local limitToFieldFilter = nil
                                    if filterData.limitToField then
                                        limitToFieldFilter = filter2
                                    end
                                    filter1:setValueCompareParams(DensityValueCompareType.BETWEEN, filterData.minState, filterData.maxState)
                                    multiModifier:addExecuteSet(filterData.toState, modifier, filter1, limitToFieldFilter)
                                end
                            end
                        end
                    end
                end
                THDensityMapConversionUtil.isDestructionInitialized = true
            end
        end)
    end
    return superFunc(sx, sz, wx, wz, hx, hz, ...)
end
function THDensityMapConversionUtil.xtHook_cutFruitArea(superFunc, extraArgs, fruitTypeIndex, sx, sz, wx, wz, hx, hz, destroySpray, useMinForageState, excludedSprayType, ...)
    local mission = g_thConversionSystem.mission
    local fieldGroundSystem = mission.fieldGroundSystem
    extraArgs = extraArgs or THDensityMapConversionUtil.extraFuncArgs.cutFruitArea
    local fruitTypeConversionData = nil
    local modifierData, usedGrowthState = nil, -1
    local outputFillType = FillType.UNKNOWN
    local vehicle = extraArgs.vehicle
    local vehicleName = "Unknown"
    if vehicle ~= nil and vehicle.getName ~= nil then
        vehicleName = vehicle:getName()
    end
    local allowCutting = true
    local prependSuccess = false
    if extraArgs.isActive and vehicle ~= nil
        and fruitTypeIndex ~= nil and fruitTypeIndex ~= FruitType.UNKNOWN
    then
        THUtils.call(function()
            fruitTypeConversionData = g_thConversionSystem:getFruitTypeConversionData(fruitTypeIndex)
            if fruitTypeConversionData ~= nil
                and (extraArgs.inputFruitType == nil
                    or extraArgs.inputFruitType == fruitTypeConversionData.fruitType.index)
            then
                local fruitTypeData = fruitTypeConversionData.fruitType.data
                local fruitTypeInfo = fruitTypeConversionData.fruitType.info
                local supportedFillTypes = extraArgs.supportedFillTypes
                local requiredFillType = extraArgs.requiredFillType or FillType.UNKNOWN
                local inputFillType = extraArgs.inputFillType
                if inputFillType == nil or inputFillType == FillType.UNKNOWN then
                    inputFillType = fruitTypeData.fillTypeIndex
                    extraArgs.inputFillType = inputFillType
                end
                if extraArgs.modifierIndex ~= nil and extraArgs.modifierIndex > 0 then
                    modifierData = g_thConversionSystem:getFruitTypeModifierDataByIndex(fruitTypeIndex, extraArgs.modifierIndex)
                    usedGrowthState = fruitTypeData.minHarvestState
                    if useMinForageState then
                        usedGrowthState = fruitTypeData.minForageState
                    end
                else
                    modifierData, usedGrowthState = THDensityMapConversionUtil.getFruitTypeAreaModifiers(fruitTypeIndex, sx, sz, wx, wz, hx, hz, useMinForageState, extraArgs)
                end
                outputFillType = inputFillType
                if usedGrowthState > 0 then
                    if modifierData ~= nil then
                        extraArgs.modifierIndex = modifierData.index
                        if modifierData.values.toFillType ~= nil then
                            outputFillType = modifierData.values.toFillType
                        end
                    end
                    if outputFillType == nil or outputFillType == FillType.UNKNOWN
                        or (supportedFillTypes ~= nil and not supportedFillTypes[outputFillType])
                        or (requiredFillType ~= FillType.UNKNOWN and outputFillType ~= requiredFillType)
                    then
                        allowCutting = false
                        if THDebugUtil.getIsEnabled(debugFlagId) then
                            local outputFillTypeName = g_thFillTypeManager:getFillTypeName(outputFillType)
                            THUtils.displayMsg("%s >> Cannot cut fruitType [%s] (%s)", vehicleName, fruitTypeData.index, fruitTypeData.name)
                            THUtils.displayMsg(" - Output fillType [%s] (%s) not supported", outputFillType, outputFillTypeName)
                            THUtils.displayMsg("")
                        end
                    end
                    if allowCutting then
                        if modifierData ~= nil then
                            local modifierValues = modifierData.values
                            local conditionData = modifierData.conditions
                            if destroySpray and modifierValues.toChopperType ~= nil then
                                local oldChopperValue = nil
                                if fruitTypeInfo.chopperTypeIndex ~= nil then
                                    oldChopperValue = fieldGroundSystem:getChopperTypeValue(fruitTypeInfo.chopperTypeIndex)
                                end
                                if excludedSprayType == nil or excludedSprayType == oldChopperValue then
                                    excludedSprayType = modifierValues.toChopperValue
                                end
                            end
                            local toHarvestState = modifierValues.toHarvestState
                            local minHarvestState = fruitTypeData.minHarvestState
                            local maxHarvestState = fruitTypeData.maxHarvestState
                            if toHarvestState ~= nil and toHarvestState < minHarvestState then
                                useMinForageState = true
                            end
                            if useMinForageState then
                                minHarvestState = fruitTypeData.minForageState
                            end
                            if usedGrowthState >= minHarvestState and usedGrowthState <= maxHarvestState then
                                toHarvestState = usedGrowthState
                            else
                                toHarvestState = minHarvestState
                            end
                            for growthState in pairs(conditionData.allowedStates) do
                                if (growthState < minHarvestState or growthState > maxHarvestState)
                                    and growthState ~= toHarvestState
                                then
                                    THDensityMapConversionUtil.convertCutFruitArea(fruitTypeIndex, growthState, sx, sz, wx, wz, hx, hz, nil, toHarvestState)
                                end
                            end
                        end
                        prependSuccess = true
                    end
                end
            end
        end)
    end
    local function appendFunc(rChangedArea, rRealArea, ...)
        THUtils.call(function()
            if prependSuccess -- implies other checks
                and rChangedArea ~= nil and rChangedArea > 0
                and fruitTypeConversionData ~= nil
            then
                local fruitTypeInfo = fruitTypeConversionData.fruitType.info
                if modifierData ~= nil then
                    local modifierValues = modifierData.values
                    local cutState = fruitTypeInfo.cutState or -1
                    if cutState >= 0 then
                        local toFruitTypeIndex = modifierValues.toFruitType
                        local toCutState = modifierValues.toCutState
                        if toFruitTypeIndex ~= nil and toCutState == nil then
                            local toFruitTypeInfo = g_thFruitTypeManager:getFruitType(toFruitTypeIndex)
                            if toFruitTypeInfo ~= nil then
                                toCutState = toFruitTypeInfo.cutState or 0
                            end
                        end
                        local toGroundType = modifierValues.toGroundType
                        local toChopperType = modifierValues.toChopperType
                        local removeHaulm = modifierValues.removeHaulm or false
                        THDensityMapConversionUtil.convertCutFruitArea(fruitTypeIndex, cutState, sx, sz, wx, wz, hx, hz, toFruitTypeIndex, toCutState, toGroundType, toChopperType, removeHaulm)
                    end
                end
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local inputFillType = extraArgs.inputFillType or FillType.UNKNOWN
                    local inputFillTypeName = g_thFillTypeManager:getFillTypeName(inputFillType)
                    local outputFillTypeName = g_thFillTypeManager:getFillTypeName(outputFillType)
                    THUtils.displayMsg("%s >> Harvest parameters for fruitType [%s] (%s):", vehicleName, fruitTypeInfo.index, fruitTypeInfo.name)
                    THUtils.displayMsg("- Input fillType: [%s] %s", inputFillType, inputFillTypeName)
                    THUtils.displayMsg("- Output fillType: [%s] %s", outputFillType, outputFillTypeName)
                    THUtils.displayMsg("- Growth state: %s", usedGrowthState)
                    THUtils.printTable(extraArgs)
                    if modifierData ~= nil then
                        THUtils.displayMsg("- Modifier values:")
                        THUtils.printTable(modifierData.values)
                        THUtils.displayMsg("- Modifier conditions:")
                        THUtils.printTable(modifierData.conditions, 1)
                    end
                end
            end
            THUtils.clearTable(THDensityMapConversionUtil.extraFuncArgs.cutFruitArea)
        end)
        return rChangedArea, rRealArea, ...
    end
    if allowCutting then
        return appendFunc(superFunc(fruitTypeIndex, sx, sz, wx, wz, hx, hz, destroySpray, useMinForageState, excludedSprayType, ...))
    end
    return appendFunc(0, 0)
end
g_thConversionSystem:addDelayedLoadTask(function()
    THUtils.setFunctionHook("FSDensityMapUtil", "updateWheelDestructionArea", THDensityMapConversionUtil, nil, false)
end)